
#include <errno.h>
#include <stdio.h>

#include <moep/types.h>
#include <moep/ieee80211_addr.h>

#include "list.h"
#include "timeout.h"
#include "neighbor.h"

#define DEFAULT_NB_TIMEOUT 1000*10

struct neighbor {
	struct list_head list;
	u8 hwaddr[IEEE80211_ALEN];
	u16 lseq;
	timeout_t timeout;
};


LIST_HEAD(nl); // neighbor list


/*
 * Callback for timeout setup in Problem 1b). The function is not called
 * is called automatically when the timeout expires. The argument data points
 * to the neighbor in question. The other two arguments are unused.
 */
static int cb_delete(timeout_t t, u32 overrun, void *data)
{
	(void) overrun;
	(void) t;
	struct neighbor *nb = (struct neighbor *)data;
	nb_del(nb->hwaddr);
	return 0;
}

/*
 * Iterate over the neighbor list nb and find the neighbor using the supplied
 * hwaddr. If the neighbor is found, return a pointer to it, and NULL
 * otherwise.
 */
static struct neighbor *find(const u8 *hwaddr)
{
	struct neighbor *cur;

	list_for_each_entry(cur, &nl, list) {
		if (0 == memcmp(hwaddr, cur->hwaddr, IEEE80211_ALEN))
			return cur;
	}
	return NULL;
}

/*
 * Create a new neighbor from the given hwaddr. Return 0 on success and -1
 * with errno set appropriately otherwise, e.g. EEXIST if the neighbor in
 * question already exists.
 */
int nb_add(const u8 *hwaddr)
{
	struct neighbor *nb;

	// Check if the neighbor already exists
	if (find(hwaddr)) {
		errno = EEXIST;
		return -1;
	}

	// Create a neighbor
	nb = malloc(sizeof(struct neighbor));
	memset(nb, 0, sizeof(*nb));
	nb->lseq = 65535;

	memcpy(nb->hwaddr, hwaddr, IEEE80211_ALEN);

	if (0 > timeout_create(CLOCK_MONOTONIC, &nb->timeout, cb_delete, nb)) {
		fprintf(stderr, "timeout_create() failed\n");
		return -1;
	}

	timeout_settime(nb->timeout, 0, timeout_msec(DEFAULT_NB_TIMEOUT, 0));

	// Add the neighbor to the list
	list_add(&nb->list, &nl);

	fprintf(stderr, "new neighbor %s\n", ieee80211_ntoa(hwaddr));

	return 0;
}

/*
 * Delete a neighbor given its hwaddr. Return 0 on success and -1 otherwise
 * with errno set appropriately.
 */
int nb_del(const u8 *hwaddr)
{
	struct neighbor *nb;

	if (!(nb = find(hwaddr))) {
		errno = EFAULT;
		return -1;
	}

	fprintf(stderr, "deleting neighbor %s\n", ieee80211_ntoa(hwaddr));

	list_del(&nb->list);
	timeout_delete(nb->timeout);
	free(nb);

	return 0;
}

/*
 * Called whenever a frame is received. hwaddr is the transmitter (neighbor
 * address), seq is the sequence number of the frame.
 * Find the neighbor given hwaddr or create a new neighbor if neccessary.
 * Set / update the neighbor's lseq (last sequence number).
 * In Problem 1b) add also update the neighbor's timeout. In Problem 2)
 * additional code for updating link qualities may be added.
 */
int nb_update_seq(const u8 *hwaddr, u16 seq)
{
	struct neighbor *nb;

	if (0 > nb_add(hwaddr)) {
		if (errno != EEXIST)
			return -1;
	}

	if (!(nb = find(hwaddr))) {
		// This must never happen
		return -1;
	}

	nb->lseq = seq;

	return timeout_settime(nb->timeout, 0, timeout_msec(DEFAULT_NB_TIMEOUT, 0));
}
